from collections import OrderedDict
import hug.types


def generate(module, base_url=""):
    documentation = OrderedDict()
    overview = module.__doc__
    if overview:
        documentation['overview'] = overview
    for url, method_handler in module.HUG_API_CALLS.items():
        url_doc = documentation.setdefault(url, {})

        mapping = OrderedDict()
        for method, handler in method_handler.items():
            mapping.setdefault(handler, []).append(method.split("_")[-1].upper())

        for handler, methods in mapping.items():
            doc = url_doc.setdefault(",".join(methods), OrderedDict())
            usage = handler.api_function.__doc__
            if usage:
                doc['usage'] = usage
            if handler.example:
                doc['example'] = "{0}{1}".format(base_url, url)
                if isinstance(handler.example, str):
                    doc['example'] += "?{0}".format(handler.example)
            doc['outputs'] = OrderedDict(format=handler.output_format.__doc__)

            if handler.accepted_parameters:
                inputs = doc.setdefault('inputs', OrderedDict())
                types = handler.api_function.__annotations__
                for argument in handler.accepted_parameters:
                    if argument in ('request', 'response'):
                        continue

                    input_definition = inputs.setdefault(argument, OrderedDict())
                    input_definition['type'] = types.get(argument, hug.types.text).__doc__
                    default = handler.defaults.get(argument, None)
                    if default is not None:
                        input_definition['default'] = default

    return documentation




import json as json_converter
from datetime import datetime


def _json_converter(item):
    if isinstance(item, datetime):
        return item.isoformat()
    elif isinstance(item, bytes):
        return item.decode('utf8')
    raise TypeError("Type not serializable")


def json(content, **kwargs):
    """JSON (Javascript Serialized Object Notation)"""
    return json_converter.dumps(content, default=_json_converter, **kwargs).encode('utf8')


def text(content):
    """Free form UTF8 text"""
    return content.encode('utf8')


def _camelcase(dictionary):
    if not isinstance(dictionary, dict):
        return dictionary

    new_dictionary = {}
    for key, value in dictionary.items():
        if isinstance(key, str):
            key = key[0] + "".join(key.title().split('_'))[1:]
        new_dictionary[key] = _camelcase(dictionary)


def json_camelcase(content):
    """JSON (Javascript Serialized Object Notation) with all keys camelCased"""
    return _camelcase(json(body))


def pretty_json(content):
    """JSON (Javascript Serialized Object Notion) pretty printed and indented"""
    return json(content, indent=4, separators=(',', ': '))



from functools import wraps, partial
from collections import OrderedDict
import sys
from hug.run import server
import hug.output_format

from falcon import HTTP_METHODS, HTTP_BAD_REQUEST


def call(urls=None, accept=HTTP_METHODS, output=hug.output_format.json, example=None):
    if isinstance(urls, str):
        urls = (urls, )

    def decorator(api_function):
        module = sys.modules[api_function.__module__]
        accepted_parameters = api_function.__code__.co_varnames[:api_function.__code__.co_argcount]
        takes_kwargs = bool(api_function.__code__.co_flags & 0x08)

        defaults = {}
        for index, default in enumerate(reversed(api_function.__defaults__ or ())):
            defaults[accepted_parameters[-(index + 1)]] = default
        required = accepted_parameters[:-(len(api_function.__defaults__ or ())) or None]
        use_example = example
        if not required and example is None:
            use_example = True

        def interface(request, response):
            input_parameters = request.params
            errors = {}
            for key, type_handler in api_function.__annotations__.items():
                try:
                    if key in input_parameters:
                        input_parameters[key] = type_handler(input_parameters[key])
                except Exception as error:
                    errors[key] = str(error)

            input_parameters['request'], input_parameters['response'] = (request, response)
            for require in required:
                if not require in input_parameters:
                    errors[require] = "Required parameter not supplied"
            if errors:
                response.data = output({"errors": errors})
                response.status = HTTP_BAD_REQUEST
                return

            if not takes_kwargs:
                input_parameters = {key: value for key, value in input_parameters.items() if key in accepted_parameters}

            response.data = output(api_function(**input_parameters))

        if not 'HUG' in module.__dict__:
            def api_auto_instantiate(*kargs, **kwargs):
                module.HUG = server(module)
                return module.HUG(*kargs, **kwargs)
            module.HUG = api_auto_instantiate
            module.HUG_API_CALLS = OrderedDict()

        for url in urls or ("/{0}".format(api_function.__name__), ):
            handlers = module.HUG_API_CALLS.setdefault(url, {})
            for method in accept:
                handlers["on_{0}".format(method.lower())] = interface

        api_function.interface = interface
        interface.api_function = api_function
        interface.output_format = output
        interface.example = use_example
        interface.defaults = defaults
        interface.accepted_parameters = accepted_parameters

        return api_function
    return decorator


for method in HTTP_METHODS:
    globals()[method.lower()] = partial(call, accept=(method, ))
